/* Copyright (C) 1999 Chris Vine, G3XXF

This program is distributed under the General Public Licence, version 2.
For particulars of this and relevant disclaimers see the file
COPYRIGHT distributed with the source files.

*/

// keyboard.cpp

#include "keyboard.h"

MainScreen* Keyboard_input_with_data::mainscreen_ptr;
SendWin* Keyboard_input_with_data::sendwin_ptr;
ReceiveWin* Keyboard_input_with_data::receivewin_ptr;
WinList* Keyboard_input_with_data::screenform_ptr;
Pipe_fifo* Keyboard_input_with_data::send_pipe_ptr;
Transmit_buffer* Keyboard_input_with_data::tr_buffer_ptr;
Keyboard_input* Keyboard_input_with_data::default_input_ptr;
Tnc* Keyboard_input_with_data::tnc_ptr;
Keyboard_input* Keyboard_input_with_data::keyboard_ptr;
int Keyboard_input_with_data::sendfile_flag;
int Keyboard_input_with_data::convert_codes_flag;

// the following matrix contains the assigned key representing numbers, for 3 series
// of keys F1 to F4 (one for linux console, one for xterm and one for any user keycode
// supplied with F1_KEYCODE (passed by the Makefile))

int Keyboard_input_with_data::F_KEYmatrix[3][7];

// now we need to define some key codes for F1 to F4 in case terminfo does not define them
// properly, and to define some key codes for the keys pressed with Alt

// check first that ncurses is of a sufficiently recent version
#if NCURSES_VERSION_PATCH < 970816
#error :  ncurses 4.2 or above must be installed.
#endif


const int ADDED_FKEY0 = 1000;
inline int ADDED_FKEY(unsigned char n) {
    if (n > 0 && n < 5) return (n - 1);
    else if (n > 9 && n < 13) return (n - 6);
    else return 0;
}

const char LINUX_F1[] = "\e[[A";
const char XTERM_F1[] = "\e[11~";
const char F10[] = "\e[21~";
const char F11[] = "\e[23~";
const char F12[] = "\e[24~";

// the following function is passed to the constructor of the first object declared and deriving from
// the Screen class in screen.h

void curses_init(void) {
// set up basic keyboard functions
    keypad(stdscr, TRUE);
    noecho();
    cbreak();
    nodelay(stdscr, TRUE);
    raw();
    nl();

    int index;
    char keycode[10];

// initialise F_KEYmatrix
    int keynumber = ADDED_FKEY0 + 1;
    int index2;
    for (index = 0; index < 3; index++) {
        for (index2 = 0; index2 < 7; index2++) {
	    Keyboard_input_with_data::F_KEYmatrix[index][index2] = keynumber;
	    keynumber++;
	}
    }

// now define the keycodes for F1 - F4
    strcpy(keycode, LINUX_F1);
    for (index = 0; index < 4; index++, keycode[3]++) {
        define_key(keycode, Keyboard_input_with_data::F_KEYmatrix[0][index]);
    }
    strcpy(keycode, XTERM_F1);
    for (index = 0; index < 4; index++, keycode[3]++) {
        define_key(keycode, Keyboard_input_with_data::F_KEYmatrix[1][index]);
    }
    if (strlen(F1_USERCODE) && strlen(F1_USERCODE) < 8) {
        strcpy(keycode, "\e");
	strcat(keycode, F1_USERCODE);
// find the last alphanumeric
	char* letter = keycode + strlen(keycode);
	while (!isalnum((unsigned char)*letter) && letter > keycode) letter--;
	if (letter > keycode) {
	    for (index = 0; index < 4; index++, (*letter)++) {
	        define_key(keycode, Keyboard_input_with_data::F_KEYmatrix[2][index]);
	    }
	}
    }
// now define default keycodes for Linux console and xterm for F10 - F12
    define_key((char*)F10, Keyboard_input_with_data::F_KEYmatrix[0][ADDED_FKEY(10)]);
    define_key((char*)F11, Keyboard_input_with_data::F_KEYmatrix[0][ADDED_FKEY(11)]);
    define_key((char*)F12, Keyboard_input_with_data::F_KEYmatrix[0][ADDED_FKEY(12)]);
}

int Keyboard_input::input_confirm(void) {
  // this returns -1 if there is not valid input,
  // or 1 if the answer is yes, or 0 if the answer is no
    int result = -1;
    int letter = getch();
    if (letter == 'y' || letter == 'Y') result = 1;
    if (letter == 'n' || letter == 'N') result = 0;
    return result;
}

void Keyboard_input_with_data::set_data(MainScreen* a, SendWin* b, ReceiveWin* c, WinList* d,
			        Pipe_fifo* e, Transmit_buffer* f, Tnc* g, Keyboard_input* h) {
    mainscreen_ptr = a;
    sendwin_ptr = b;
    receivewin_ptr = c;
    screenform_ptr = d;
    send_pipe_ptr = e;
    tr_buffer_ptr = f;
    tnc_ptr = g;
    default_input_ptr = h;
}

void Exit_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) prog_func.exitflag = TRUE;
    }
}

void Exit_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in exit_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Close down (y/n)?", 1, 11);
    popupwin_ptr->winrefresh();
}


void Help_file::setup(void) {
    receivewin_ptr->display_helpfile();
    keyboard_ptr = this;
}

void Help_file::action(void) {
    int letter = getch();
    if (letter == KEY_F(1)  || letter == F_KEYmatrix[0][ADDED_FKEY(1)]
		    || letter == F_KEYmatrix[1][ADDED_FKEY(1)]
		    || letter == F_KEYmatrix[2][ADDED_FKEY(1)]) {
        keyboard_ptr = default_input_ptr;
	receivewin_ptr->end_helpfile();
	refresh_winlist(*screenform_ptr);
	receivewin_ptr->redisplay();
    }
    else if (letter ==  KEY_UP) {
        receivewin_ptr->scrollup_helpwin();
    }

    else if (letter == KEY_DOWN) {
        receivewin_ptr->scrolldown_helpwin();
    }
}

void Openfile::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in Openfile::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->set_cursor_status(Screen::on);
    popupwin_ptr->write("File to be sent? (^Q to quit)", 1, 4);
    popupwin_ptr->write("", 2, 4);
    index = 0;
    if (strlen(prog_func.filedir) < 254) {
	for (char* temp_ptr = prog_func.filedir; *temp_ptr != 0 && index < 255; temp_ptr++, index++) {
	    filename[index] = *temp_ptr;
	    popupwin_ptr->write(filename[index]);
	}
	filename[index] = '/';
	popupwin_ptr->write('/');
	index++;
    }
    popupwin_ptr->winrefresh();
}

void Openfile::action(void) {
    int letter = getch();
    if (letter > 044 && letter < 123) {
        filename[index] = letter;
	index++;
	popupwin_ptr->write(letter);
	popupwin_ptr->winrefresh();
    }
    else if (letter == KEY_BACKSPACE && index) {
        index--;
	popupwin_ptr->write(8);
	popupwin_ptr->write(32);
	popupwin_ptr->write(8);
	popupwin_ptr->winrefresh();
    }
    if ((letter == '\n' && index > 0) || index == 255) {
        filename[index] = 0;
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	sendfile(filename);
    }
    if (letter == 17) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
    }
}

void Input_command::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in input_command::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->set_cursor_status(Screen::on);
    popupwin_ptr->write("Command? (^Q to quit)", 1, 6);
    popupwin_ptr->write("", 2, 6);
    popupwin_ptr->winrefresh();
    index = 0;
}

void Input_command::action(void) {
    int letter = getch();
    if (letter == ' ' || (letter > 044 && letter < 127)) {
        buffer[index] = letter;
	index++;
	popupwin_ptr->write(letter);
	popupwin_ptr->winrefresh();
    }
    else if (letter == KEY_BACKSPACE && index){
        index--;
	popupwin_ptr->write(8);
	popupwin_ptr->write(32);
	popupwin_ptr->write(8);
	popupwin_ptr->winrefresh();
    }
    if ((letter == '\n' && index > 0) || index == 254) {
        tr_buffer_ptr->reset();
	buffer[index] = 0;

	Keyboard_controller::send_kamcommand(buffer);
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
    }
    
    if (letter == 17) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
    }
}

void Callsign::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in callsign::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->set_cursor_status(Screen::on);
    popupwin_ptr->write("Callsign ? (^Q to quit)", 1, 9);
    popupwin_ptr->write("", 2, 9);
    popupwin_ptr->winrefresh();
    index = 0;
}

void Callsign::action(void) {
    int letter = getch();
    if (letter > 96 && letter < 123) letter &= 0xdf;        // convert lower to upper case
    if (letter == '-' || (letter > 46 && letter < 58)
	  || (letter > 64 && letter < 91)) {
        buffer[index] = letter;
	index++;
	popupwin_ptr->write(letter);
	popupwin_ptr->winrefresh();
    }
    else if (letter == KEY_BACKSPACE && index){
        index--;
	popupwin_ptr->write(8);
	popupwin_ptr->write(32);
	popupwin_ptr->write(8);
	popupwin_ptr->winrefresh();
    }
    if ((letter == '\n' && index > 2) || index == hisCall_SIZE) {
	buffer[index] = 0;
	strcpy(tnc_ptr->tnc_func.hisCall[tnc_ptr->tnc_func.active_stream()][tnc_ptr->tnc_func.active_port], buffer);
	if (tnc_ptr->tnc_func.active_port && tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	    tnc_ptr->tnc_func.hisCall_lock = Tnc_func::on;
	    if (tnc_ptr->is_validcall(buffer)) tnc_ptr->make_selcall(buffer, tnc_ptr->tnc_func.selCall);
	}
	mainscreen_ptr->display_callsign();
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
    }
    
    if (letter == 17) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
    }
}

void Send_parms_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    send_parms(*screenform_ptr);
	}
    }
}

void Send_parms_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in send_parms_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Send parameters to TNC (y/n)?", 1, 6);
    popupwin_ptr->winrefresh();
}

void Connect::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in connect::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->set_cursor_status(Screen::on);
    popupwin_ptr->write("Connect with ? (^Q to quit)", 1, 8);
    popupwin_ptr->write("", 2, 8);
    index = 0;
    char* temp_ptr;
    if (tnc_ptr->tnc_func.active_port && (tnc_ptr->tnc_func.hfmode == Tnc_func::amtor
      || tnc_ptr->tnc_func.hfmode == Tnc_func::lamtor
      || tnc_ptr->tnc_func.hfmode == Tnc_func::fec)) {
        temp_ptr = tnc_ptr->tnc_func.selCall;
	max_size = selCall_SIZE;
    }
    else {
        temp_ptr = tnc_ptr->tnc_func.hisCall[tnc_ptr->tnc_func.active_stream()][tnc_ptr->tnc_func.active_port];
	if (!tnc_ptr->tnc_func.active_port || tnc_ptr->tnc_func.hfmode == Tnc_func::packet) max_size = CONNECT_SIZE;
	else max_size = hisCall_SIZE;
    }
    
    for (; *temp_ptr != 0; temp_ptr++, index++) {
        buffer[index] = *temp_ptr;
	popupwin_ptr->write(*temp_ptr);
    }
    popupwin_ptr->winrefresh();
}

void Connect::action(void) {
    int letter = getch();
    if (letter > 96 && letter < 123) letter &= 0xdf;        // convert lower to upper case
    if (letter == '-' || (letter > 46 && letter < 58)
	  || (letter > 64 && letter < 91)
	  || (letter == '!' && !index && tnc_ptr->tnc_func.active_port && 
	        tnc_ptr->tnc_func.hfmode == Tnc_func::pactor)
	  || ((letter == ',' || letter == ' ') && index && 
	        (!tnc_ptr->tnc_func.active_port || tnc_ptr->tnc_func.hfmode == Tnc_func::packet))) {
        buffer[index] = letter;
	index++;
	popupwin_ptr->write(letter);
	popupwin_ptr->winrefresh();
    }
    else if (letter == KEY_BACKSPACE && index){
        index--;
	popupwin_ptr->write(8);
	popupwin_ptr->write(32);
	popupwin_ptr->write(8);
	popupwin_ptr->winrefresh();
    }
    if ((letter == '\n' && index > 2) || index == max_size) {
        tr_buffer_ptr->reset();
	buffer[index] = 0;

	if (tnc_ptr->tnc_func.active_port && tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	    Keyboard_controller::send_specialcommand(packetCMD);
	}

	char* commandmessage;
	if ((commandmessage = new char[max_size + 10]) == 0) {
	    cerr << "Memory allocation error in connect::action()" << endl;
	    exit(MEM_ERROR);
	}

	if (!tnc_ptr->tnc_func.active_port || tnc_ptr->tnc_func.hfmode == Tnc_func::packet) {
	    strcpy (commandmessage, packet_connectCMD);
	}

	else if (tnc_ptr->tnc_func.hfmode == Tnc_func::amtor ||
		   tnc_ptr->tnc_func.hfmode == Tnc_func::lamtor || 
		   tnc_ptr->tnc_func.hfmode == Tnc_func::fec) {
	    strcpy (commandmessage, amtor_connectCMD);
	    tnc_ptr->tnc_func.hfmode = Tnc_func::amtor;
	    mainscreen_ptr->display_mode();
	}

	else if (tnc_ptr->tnc_func.hfmode == Tnc_func::gtor ||
		   tnc_ptr->tnc_func.hfmode == Tnc_func::gmon) {
	    strcpy (commandmessage, gtor_connectCMD);
	    tnc_ptr->tnc_func.hfmode = Tnc_func::gtor;
	    mainscreen_ptr->display_mode();
	}

	else if (tnc_ptr->tnc_func.hfmode == Tnc_func::pactor) {
	    strcpy (commandmessage, pactor_connectCMD);
	}

	strcat(commandmessage, buffer);
	Keyboard_controller::send_kamcommand(commandmessage);

	if (!tnc_ptr->tnc_func.active_port || (tnc_ptr->tnc_func.hfmode != Tnc_func::amtor
	      && tnc_ptr->tnc_func.hfmode != Tnc_func::lamtor
	      && tnc_ptr->tnc_func.hfmode != Tnc_func::fec)) {
	    char* temp = buffer;
	    if (*temp == '!') temp++; // remove leading '!' if a pactor connect on long path
	    char* temp2;
	    int count;
	    for (count = 0, temp2 = temp; count < hisCall_SIZE
		     && *temp2 && *temp2 != ' '; temp2++, count++) {}
	    *temp2 = 0;    // terminate string at first space
	    strcpy(tnc_ptr->tnc_func.hisCall[tnc_ptr->tnc_func.active_stream()][tnc_ptr->tnc_func.active_port], temp);
	    if (tnc_ptr->tnc_func.active_port && tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        tnc_ptr->tnc_func.hisCall_lock = Tnc_func::on;
		if (tnc_ptr->is_validcall(temp)) tnc_ptr->make_selcall(temp, tnc_ptr->tnc_func.selCall);
	    }
	    mainscreen_ptr->display_callsign();
	}
	else if (tnc_ptr->tnc_func.active_port) strcpy(tnc_ptr->tnc_func.selCall, buffer);

	receivewin_ptr->redisplay(); // go to end of current active window
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	delete[] commandmessage;
    }
    
    if (letter == 17) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
    }
}


void Pactor_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(pactorCMD);

	    tnc_ptr->tnc_func.hfmode = Tnc_func::pactor;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Pactor_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in pactor_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Pactor (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}

void Amtor_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(amtorCMD);
    
	    tnc_ptr->tnc_func.hfmode = Tnc_func::amtor;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Amtor_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in amtor_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Amtor (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}


void Ascii_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(asciiCMD);

	    tnc_ptr->tnc_func.hfmode = Tnc_func::ascii;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Ascii_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in ascii_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Ascii (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}

void Fec_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(fecCMD);
   
	    tnc_ptr->tnc_func.hfmode = Tnc_func::fec;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Fec_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in fec_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Amtor FEC (y/n)?", 1, 4);
    popupwin_ptr->winrefresh();
}

void Gmon_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(gmonCMD);
	    
	    tnc_ptr->tnc_func.hfmode = Tnc_func::gmon;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Gmon_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in gmon_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Gmon (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}

void Gtor_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(gtorCMD);
	    
	    tnc_ptr->tnc_func.hfmode = Tnc_func::gtor;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Gtor_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in gtor_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Gtor (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}


void Lamtor_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(lamtorCMD);
	    
	    tnc_ptr->tnc_func.hfmode = Tnc_func::lamtor;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Lamtor_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in lamtor_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Amtor-Listen (y/n)?", 1, 4);
    popupwin_ptr->winrefresh();
}

void Packet_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    receivewin_ptr->redisplay(); // switch to end of current hf window

	    Keyboard_controller::send_specialcommand(packetCMD);
	    tnc_ptr->tnc_func.hfmode = Tnc_func::packet;
	    tnc_ptr->tnc_func.hisCall_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_callsign();
	    mainscreen_ptr->display_capture_status();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Packet_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in packet_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to Packet (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}

void Rtty_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(rttyCMD);
	    
	    tnc_ptr->tnc_func.hfmode = Tnc_func::rtty;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Rtty_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in rtty_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to RTTY (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}

void Tor_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();
	    tnc_ptr->tnc_func.active_port = 1;
	    tnc_ptr->tnc_func.active_hf_stream = 0;
	    receivewin_ptr->redisplay(); // switch current active window to hf port

	    if (tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	        Keyboard_controller::send_specialcommand(packetCMD);
	    }

	    else if (tnc_ptr->tnc_func.capture_stream.port) {// if file capture was set to HF packet, cancel
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
		mainscreen_ptr->display_capture_status();
	    }

	    Keyboard_controller::send_kamcommand(torCMD);
	    
	    tnc_ptr->tnc_func.hfmode = Tnc_func::tor;
	    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
	    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
	    mainscreen_ptr->display_mode();
	    mainscreen_ptr->display_current_stream();
	    mainscreen_ptr->display_connected_status();
	    mainscreen_ptr->make_torline();
	}
    }
}

void Tor_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in tor_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Change to TOR (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}

void Disconnect_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) {
	    tr_buffer_ptr->reset();

	    if (!tnc_ptr->tnc_func.active_port || tnc_ptr->tnc_func.hfmode == Tnc_func::packet) {
	        Keyboard_controller::send_kamcommand(packet_disconnectCMD);
	    }
	   
	    else Keyboard_controller::send_specialcommand(tor_disconnectCMD);
	}
    }
}

void Disconnect_query::setup(void) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in Disconnect_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Disconnect/abort (y/n)?", 1, 8);
    popupwin_ptr->winrefresh();
}

void Send_message_query::setup(char msg_num) {
    keyboard_ptr = this;
    popupwin_ptr = new PopupWin;
    if (!popupwin_ptr) {
        cerr << "Memory allocation error in Send_message_file_query::setup()" << endl;
	exit(MEM_ERROR);
    }
    popupwin_ptr->show();
    popupwin_ptr->write("Send message number ", 1, 6);
    popupwin_ptr->write(msg_num);
    popupwin_ptr->write(" (y/n)?");
    popupwin_ptr->winrefresh();
    message_num = msg_num;
}

void Send_message_query::action(void) {
    int result = input_confirm();
    if (result != -1) {
	keyboard_ptr = default_input_ptr;
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	if (result == 1) send_message(message_num);
    }
}

Keyboard_controller::Keyboard_controller(MainScreen& mainscreen, SendWin& sendwin,
                ReceiveWin& receivewin, WinList& screenform, Pipe_fifo& send_pipe,
		Transmit_buffer& tr_buffer, Tnc& tnc) {
    set_data(&mainscreen, &sendwin, &receivewin, &screenform, &send_pipe,
	         &tr_buffer, &tnc, this);
    sendfile_flag = not_sending;
    convert_codes_flag = FALSE;
    keyboard_ptr = this;
}

void Keyboard_controller::action(void) {
    int result;
    int termchar;

    if (sendfile_flag != not_sending) sendfile();
    if ((termchar = getch()) != ERR) {  // non-blocking getch() returns ERR if no letter available to extract
        if (sendfile_flag != loading_buffer) {

	    if (tnc_ptr->tnc_func.active_port
		&& (tnc_ptr->tnc_func.hfmode == Tnc_func::rtty
		    || tnc_ptr->tnc_func.hfmode == Tnc_func::fec
		    || tnc_ptr->tnc_func.hfmode == Tnc_func::amtor
		    || (tnc_ptr->tnc_func.hfmode == Tnc_func::gtor
			 && tnc_ptr->tnc_func.stream_status[0][1] == Tnc_func::disconnected)
		    || (tnc_ptr->tnc_func.hfmode == Tnc_func::tor
			 && (tnc_ptr->tnc_func.stream_status[0][1] == Tnc_func::disconnected
			     || tnc_ptr->tnc_func.tor_connected_mode == Tnc_func::tor_amtor)))) {
	        if (termchar > 31 && termchar < 256) {
		    if (!is_baudot((char)termchar)) termchar = 0;
		    else if (tnc_ptr->tnc_func.hfmode == Tnc_func::rtty) termchar = toupper(termchar);
		}
	    }

#if CHAR_SET==CP437
	    if ((termchar < 256  && termchar > 31) || termchar == '\n') {
#elif CHAR_SET==LATIN-1
	    if ((termchar < 257 && termchar > 31) || termchar == '\n') {
#else
	    if ((termchar < 127 && termchar > 31) || termchar == '\n') {
#endif

	        result = tr_buffer_ptr->add_letter((char)termchar);
		if (result == -1) beep();
		else sendwin_ptr->write((char)result);
		sendwin_ptr->winrefresh();
	    }

	    else if (termchar == KEY_BACKSPACE) {
	        result = tr_buffer_ptr->erase_letter();
		if (!result) {
		    beep();
		    doupdate();
		}
		else {
		    if (result == -1) {  // if last letter already extracted, we need to send value 8 to the Kam
			char tx_frame[6];
			tx_frame[0] = FEND;
			tx_frame[1] = 'D';
			tx_frame[2] = tnc_ptr->tnc_func.active_port ? '2' : '1';
			
			if ((tnc_ptr->tnc_func.active_port && tnc_ptr->tnc_func.hfmode != Tnc_func::packet)
			      || tnc_ptr->tnc_func.stream_status[tnc_ptr->tnc_func.active_stream()]
			        [tnc_ptr->tnc_func.active_port] == Tnc_func::disconnected) {
			    tx_frame[3] = '0';
			}
			else tx_frame[3] = ((tnc_ptr->tnc_func.active_stream() | 0x40) + 1);
			tx_frame[4] = 8;
			tx_frame[5] = FEND;

			do {
			    result = send_pipe_ptr->write(tx_frame, 6);
			} while (result < 1);
		    }
		    sendwin_ptr->write(8);        // now erase letter from screen
		    sendwin_ptr->write(32);
		    sendwin_ptr->write(8);
		    sendwin_ptr->winrefresh();
		}
	    }
	
	    else if ((termchar == KEY_F(4) || termchar == F_KEYmatrix[0][ADDED_FKEY(4)]
		          || termchar == F_KEYmatrix[1][ADDED_FKEY(4)]
		          || termchar == F_KEYmatrix[2][ADDED_FKEY(4)])
		      && (!tnc_ptr->tnc_func.active_port
			  || (tnc_ptr->tnc_func.hfmode != Tnc_func::rtty &&  
			      tnc_ptr->tnc_func.hfmode != Tnc_func::ascii &&
			      tnc_ptr->tnc_func.hfmode != Tnc_func::tor && 
			      tnc_ptr->tnc_func.hfmode != Tnc_func::cw))) {
	        connect_obj.setup();
	    }
	
	    else if (termchar == KEY_F(6)) {   // this gets a file name to send and opens it
	        openfile_obj.setup();
	    }
	
	    else if (termchar == KEY_F(9)) {    // this sends the tnc the parameters
	        send_parms_query_obj.setup();   // in tnc.parms
	    }

	    else if (termchar == KEY_NPAGE && tnc_ptr->tnc_func.active_port 
		       && tnc_ptr->tnc_func.hfmode != Tnc_func::packet) { // Page Down
	        if (tr_buffer_ptr->letters_free() > 5) {
		    if (tnc_ptr->tnc_func.hfmode == Tnc_func::pactor
			  || (tnc_ptr->tnc_func.hfmode == Tnc_func::gtor
			      && tnc_ptr->tnc_func.stream_status[0][1] == Tnc_func::connected)) {
		        tr_buffer_ptr->add_letter('=');
			tr_buffer_ptr->add_letter('>');
		        sendwin_ptr->write("=>");
			sendwin_ptr->winrefresh();
		    }
		    tr_buffer_ptr->add_letter(CMDinbuffer);
		    tr_buffer_ptr->add_letter('E');
		    tr_buffer_ptr->add_letter(CMDinbuffer);
		}
		else {
		    beep();
		    doupdate();
		}
	    }
	
	    else if (termchar == 1 && tnc_ptr->tnc_func.hfmode != Tnc_func::amtor) { // Ctrl-A
	        amtor_query_obj.setup();
	    }

	    else if (termchar == 4 && (!tnc_ptr->tnc_func.active_port                  // Ctrl-D
		       || tnc_ptr->tnc_func.hfmode == Tnc_func::packet
		       || tnc_ptr->tnc_func.hfmode == Tnc_func::amtor
		       || tnc_ptr->tnc_func.hfmode == Tnc_func::gtor
		       || tnc_ptr->tnc_func.hfmode == Tnc_func::pactor
		       || tnc_ptr->tnc_func.hfmode == Tnc_func::tor)) {
	        disconnect_query_obj.setup();
	    }

	    else if (termchar == 6 && tnc_ptr->tnc_func.hfmode != Tnc_func::fec // Ctrl-F
		     && tnc_ptr->tnc_func.active_port) {
	        fec_query_obj.setup();
	    }

	    else if (termchar == 7 && tnc_ptr->tnc_func.hfmode != Tnc_func::gtor // Ctrl-G
		     && tnc_ptr->tnc_func.active_port) {
	        gtor_query_obj.setup();
	    }

	    else if (termchar == 12 && tnc_ptr->tnc_func.hfmode != Tnc_func::lamtor // Ctrl-L
		     && tnc_ptr->tnc_func.active_port) {
	        lamtor_query_obj.setup();
	    }

	    else if (termchar == 15 && tnc_ptr->tnc_func.hfmode != Tnc_func::gmon   // Ctrl-O
		     && tnc_ptr->tnc_func.active_port) {
	        gmon_query_obj.setup();
	    }

	    else if (termchar == 16 && tnc_ptr->tnc_func.hfmode != Tnc_func::pactor // Ctrl-P
		     && tnc_ptr->tnc_func.active_port) {
	        pactor_query_obj.setup();
	    }

	    else if (termchar == 18 && tnc_ptr->tnc_func.hfmode != Tnc_func::rtty // Ctrl-R
		     && tnc_ptr->tnc_func.active_port) {
	        rtty_query_obj.setup();
	    }

	    else if (termchar == 19 && tnc_ptr->tnc_func.hfmode != Tnc_func::ascii // Ctrl-S
		     && tnc_ptr->tnc_func.active_port) {
	        ascii_query_obj.setup();
	    }

	    else if (termchar == 20 && tnc_ptr->tnc_func.hfmode != Tnc_func::tor   // Ctrl-T
		     && tnc_ptr->tnc_func.active_port) {
	        tor_query_obj.setup();
	    }

	    else if (termchar == 24 && tnc_ptr->tnc_func.hfmode != Tnc_func::packet // Ctrl-X
		     && tnc_ptr->tnc_func.active_port) {
	        packet_query_obj.setup();
	    }
	}

	if (termchar == 17) {      // Ctrl-Q
	    exit_query_obj.setup();
	}

	if (termchar == KEY_UP) {
	    receivewin_ptr->scrollup();
	}

	else if (termchar == KEY_DOWN) {
	    receivewin_ptr->scrolldown();
	}

	else if (termchar == KEY_RIGHT) {
	    receivewin_ptr->scrollout();
	}

	else if (termchar == KEY_PPAGE && tnc_ptr->tnc_func.active_port 
		       && tnc_ptr->tnc_func.hfmode != Tnc_func::packet) { // Page Up
	    Keyboard_controller::send_specialcommand(txCMD);
	}
	    
	else if (termchar == KEY_F(1)  || termchar == F_KEYmatrix[0][ADDED_FKEY(1)]
		    || termchar == F_KEYmatrix[1][ADDED_FKEY(1)]
		    || termchar == F_KEYmatrix[2][ADDED_FKEY(1)]) {   // this displays the help file
	    help_obj.setup();
	}
	
	else if (termchar == KEY_F(2) || termchar == F_KEYmatrix[0][ADDED_FKEY(2)]
		    || termchar == F_KEYmatrix[1][ADDED_FKEY(2)]
		    || termchar == F_KEYmatrix[2][ADDED_FKEY(2)]) {   // this toggles the file capture status
	    if (tnc_ptr->tnc_func.capturefile_flag) {
	        tnc_ptr->tnc_func.capturefile_flag = FALSE;
		fileout.close();
	    }

	    else {
	        tnc_ptr->tnc_func.capturefile_flag = TRUE;
		tnc_ptr->tnc_func.capture_stream.port = tnc_ptr->tnc_func.active_port;
		tnc_ptr->tnc_func.capture_stream.stream = tnc_ptr->tnc_func.active_stream();
		char* filename = new char[strlen(prog_func.filedir) + 13];
		if (!filename) {
		    cerr << "Memory allocation error in Keyboard_controller.action()" << endl;
		    exit(MEM_ERROR);
		}
		strcpy(filename, prog_func.filedir);
		strcat(filename, "/capture.txt");

		fileout.open(filename, ios::out | ios::app);
		if (!fileout) {
		    popupwin_ptr = new PopupWin;
		    if (!popupwin_ptr) {
		        cerr << "Memory allocation error in Keyboard_controller.action()" << endl;
			exit(MEM_ERROR);
		    }
		    popupwin_ptr->show();
		    popupwin_ptr->write("File capture.txt could not be opened", 2, 5);
		    popupwin_ptr->winrefresh();
		    sleep(3);
		    popupwin_ptr->unshow();
		    delete popupwin_ptr;
		    refresh_winlist(*screenform_ptr);
		    tnc_ptr->tnc_func.capturefile_flag = FALSE;
		}
		delete[] filename;
	    }
	    mainscreen_ptr->display_capture_status();
	}

	else if (termchar == KEY_F(3) || termchar == F_KEYmatrix[0][ADDED_FKEY(3)]
		     || termchar == F_KEYmatrix[1][ADDED_FKEY(3)]
		     || termchar == F_KEYmatrix[2][ADDED_FKEY(3)]) {   // this toggles the line/guard/word sending status
	    if (prog_func.send_mode == Prog_func::line) prog_func.send_mode = Prog_func::word;
	    else if (prog_func.send_mode == Prog_func::word) prog_func.send_mode = Prog_func::guard;
	    else prog_func.send_mode = Prog_func::line;
	    mainscreen_ptr->display_send_mode_status();
	}

	else if (termchar == KEY_F(5)) {   // send a command to the Kam
	    if (tr_buffer_ptr->letters_used() < FRAMEDATA_SIZE 
		&& tnc_ptr->tnc_func.stream_status[tnc_ptr->tnc_func.active_stream()][tnc_ptr->tnc_func.active_port] == Tnc_func::disconnected) {
	        input_command_obj.setup();
	    }
	    else {
	        beep();
		doupdate();
	    }
	}
	
	else if (termchar == KEY_F(8) && tnc_ptr->tnc_func.active_port // this toggles the call lock status
		   && tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	    if (tnc_ptr->tnc_func.hisCall_lock == Tnc_func::off) tnc_ptr->tnc_func.hisCall_lock = Tnc_func::on;
	    else tnc_ptr->tnc_func.hisCall_lock = Tnc_func::off;
	    mainscreen_ptr->display_callsign();
	}

	else if (termchar == KEY_F(7)) {   // this displays enters a call in the callsign buffer
	    callsign_obj.setup();
	}

	else if (termchar == KEY_F(10)    // toggles active port between hf and vhf
		 || termchar == F_KEYmatrix[0][ADDED_FKEY(10)]) {
	    if(tr_buffer_ptr->letters_used() < FRAMEDATA_SIZE
	       && tnc_ptr->get_freebytes() > tr_buffer_ptr->letters_used() + 3) {
	        if (!tnc_ptr->tnc_func.active_port) tnc_ptr->tnc_func.active_port = 1;
		else tnc_ptr->tnc_func.active_port = 0;
		receivewin_ptr->redisplay(); // switch current active window in display
		mainscreen_ptr->display_current_stream();
		mainscreen_ptr->display_callsign();
		mainscreen_ptr->display_connected_status();
		mainscreen_ptr->make_torline();
		mainscreen_ptr->update_lockinfo();
		mainscreen_ptr->update_torinfo();
		mainscreen_ptr->update_txinfo();
                               // notify the Kam LEDs of the change
		Keyboard_controller::send_kamcommand("");
	    }
	    else {
	        beep();
		doupdate();
	    }
	}

	else if (termchar == KEY_F(11)   // goes down one packet stream on active channel
		 || termchar == F_KEYmatrix[0][ADDED_FKEY(11)]) {
	    if(tr_buffer_ptr->letters_used() < FRAMEDATA_SIZE
	       && tnc_ptr->get_freebytes() > tr_buffer_ptr->letters_used() + 3) {
	        if (tnc_ptr->tnc_func.active_port && tnc_ptr->tnc_func.hfmode == Tnc_func::packet) {
		    if (!tnc_ptr->tnc_func.active_hf_stream) tnc_ptr->tnc_func.active_hf_stream = MAXUSERS - 1;
		    else tnc_ptr->tnc_func.active_hf_stream--;
		}
		else if (!tnc_ptr->tnc_func.active_port) {
		    if (!tnc_ptr->tnc_func.active_vhf_stream) tnc_ptr->tnc_func.active_vhf_stream = MAXUSERS - 1;
		    else tnc_ptr->tnc_func.active_vhf_stream--;
		}
		else return;

		receivewin_ptr->redisplay(); // switch current active window in display
		mainscreen_ptr->display_current_stream();
		mainscreen_ptr->display_callsign();
		mainscreen_ptr->display_connected_status();

                               // notify the Kam LEDs of the change
		Keyboard_controller::send_kamcommand("");
	    }
	    else {
	        beep();
		doupdate();
	    }
	}


	else if (termchar == KEY_F(12)   // goes up one packet stream on active channel
		 || termchar == F_KEYmatrix[0][ADDED_FKEY(12)]) {
	    if(tr_buffer_ptr->letters_used() < FRAMEDATA_SIZE
	       && tnc_ptr->get_freebytes() > tr_buffer_ptr->letters_used() + 3) {
	        if (tnc_ptr->tnc_func.active_port && tnc_ptr->tnc_func.hfmode == Tnc_func::packet) {
		    tnc_ptr->tnc_func.active_hf_stream++;
		    if (tnc_ptr->tnc_func.active_hf_stream == MAXUSERS) tnc_ptr->tnc_func.active_hf_stream = 0;
		}
		else if (!tnc_ptr->tnc_func.active_port) {
		    tnc_ptr->tnc_func.active_vhf_stream++;
		    if (tnc_ptr->tnc_func.active_vhf_stream == MAXUSERS) tnc_ptr->tnc_func.active_vhf_stream = 0;
		}
		else return;
		
		receivewin_ptr->redisplay(); // switch current active window in display
		mainscreen_ptr->display_current_stream();
		mainscreen_ptr->display_callsign();
		mainscreen_ptr->display_connected_status();

                               // notify the Kam LEDs of the change
		Keyboard_controller::send_kamcommand("");
	    }	    
	    else {
	        beep();
		doupdate();
	    }
	}

	else if (termchar == 27) {   // Alt key?
	    termchar = getch();

	    if ((termchar == 'L' || termchar == 'l')
		   && tnc_ptr->tnc_func.active_port
		   && (tnc_ptr->tnc_func.hfmode == Tnc_func::pactor
		         || tnc_ptr->tnc_func.hfmode == Tnc_func::gtor)) {
	        if (tnc_ptr->tnc_func.speed_lock == Tnc_func::off) {
		    tnc_ptr->tnc_func.speed_lock = Tnc_func::on;
		    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::yes;
		    Keyboard_controller::send_specialcommand('1');
		}
		else {
		    tnc_ptr->tnc_func.speed_lock = Tnc_func::off;
		    tnc_ptr->tnc_func.have_sent_lock = Tnc_func::no;
		    Keyboard_controller::send_specialcommand('0');
		}
		mainscreen_ptr->update_lockinfo();
	    }

	    else if ((termchar == 'I' || termchar == 'i')
		        && tnc_ptr->tnc_func.active_port
		        && tnc_ptr->tnc_func.hfmode != Tnc_func::packet
		        && sendfile_flag != loading_buffer) {
	        char* message = new char[strlen(prog_func.myCall) + hisCall_SIZE + 5]; // must be large enough for
		if (!message) {
		    cerr << "Memory allocation error in Keyboard_controller::action()" << endl;
		    exit(MEM_ERROR);
		}
	        strcpy(message, tnc_ptr->tnc_func.hisCall[0][1]);     // myCall and hisCall and " de "
		if (tnc_ptr->tnc_func.hfmode == Tnc_func::rtty) strcat(message, " DE ");
		else strcat(message, " de ");
		strcat(message, prog_func.myCall);
		strcat(message, " ");

		int index;
		for (index = result = 0; result != -1 && message[index]; index++) {
		    result = tr_buffer_ptr->add_letter(message[index]);
		    if (result == -1) beep();
		    else sendwin_ptr->write((char)result);
		}
		sendwin_ptr->winrefresh();
		delete[] message;
	    }
	    else if (termchar >= '0' && termchar <= '9') {
	        if (sendfile_flag != loading_buffer) {
		    if (prog_func.noprompt_flag == FALSE) send_message_query_obj.setup((char)termchar);
		    else send_message((char)termchar);
		}
		else {
		    beep();
		    doupdate();
		}
	    }
	}
    }
}

void Keyboard_controller::send_kamcommand(const char* text, char ascii_stream) {
    int result;
    if (ascii_stream == -1) {
        if (tnc_ptr->tnc_func.active_port && tnc_ptr->tnc_func.hfmode != Tnc_func::packet) {
	    ascii_stream = '0';
	}
	else ascii_stream = ((tnc_ptr->tnc_func.active_stream() | 0x40) + 1);
    }

    char command_string[5];
    command_string[0] = FEND;
    command_string[1] = 'C';
    command_string[2] = tnc_ptr->tnc_func.active_port ? '2' : '1';
    command_string[3] = ascii_stream;
    command_string[4] = 0;

    do {
        result = send_pipe_ptr->write(command_string);
    } while (!result);


    if (*text) {  // text may be null if just the Kam LEDs are to be redirected
        do {
	    result = send_pipe_ptr->write(text);
	} while (!result);
    }

    do {
        result = send_pipe_ptr->write(FEND);
    } while (!result);
    usleep(20000);
}

void Keyboard_controller::send_specialcommand(char letter) {
    int result;
    do {
        result = send_pipe_ptr->write(FEND);
    } while (!result);

    do {
        result = send_pipe_ptr->write(letter);
    } while (!result);

    do {
        result = send_pipe_ptr->write(FEND);
    } while (!result);
    usleep(20000);
}

void Keyboard_input_with_data::send_message(char message_number) {
    char* filename = new char[strlen(prog_func.filedir) + 10];
    if (!filename) {
        cerr << "Memory allocation error in Keyboard_controller::send_message(char)" << endl;
	exit(MEM_ERROR);
    }
    strcpy(filename, prog_func.filedir);
    strcat(filename, "/messages");

    filein.open(filename, ios::in | ios::nocreate);
    if(!filein) {
        PopupWin* popupwin_ptr = new PopupWin;
	if (!popupwin_ptr) {
	    cerr << "Memory allocation error in Keyboard_controller::send_message()" << endl;
	    exit(MEM_ERROR);
	}
        popupwin_ptr->show();
	popupwin_ptr->write(filename, 1, 5);
	popupwin_ptr->write("cannot be found or opened", 2, 5);
	popupwin_ptr->winrefresh();
	sleep(3);
	popupwin_ptr->unshow();
	delete popupwin_ptr;
	refresh_winlist(*screenform_ptr);
	filein.clear();
    }
    else {
	char buffer[MESSAGE_SIZE + 1];
	int extracted = 0;
	int result = 0;
	while(!result && filein.getline(buffer, MESSAGE_SIZE)) {
	    if (*buffer == '#' && *(buffer + 1) == message_number) {
	        // now get next line
	        if (filein.getline(buffer, MESSAGE_SIZE) && (extracted = filein.gcount()) > 0)  {
		    buffer[--extracted] = 0;  // null terminate last line extracted
		    extracted++;              // make extracted equal to gcount() again
		    result = 1;
		}
	    }
	}
	if (result) {
	    if (message_number == '0' || message_number > '5') {  // send a file
	        filein.close();
		sendfile(buffer, TRUE);
	    }
	    else {  // send the message in ./messages
	        int count = 0;
		int index;
		char letter;

		char* filename = new char[strlen(prog_func.filedir) + 11];
		if (!filename) {
		    cerr << "Memory allocation error in Keyboard_controller.action()" << endl;
		    exit(MEM_ERROR);
		}
		strcpy(filename, prog_func.filedir);
		strcat(filename, "/temp_file");

		ofstream tempfile(filename, ios::out | ios::noreplace);
		if (!tempfile) {
		    PopupWin* popupwin_ptr = new PopupWin;
		    if (!popupwin_ptr) {
		        cerr << "Memory allocation error in Keyboard_controller.send_message(char)" << endl;
			exit(MEM_ERROR);
		    }
		    popupwin_ptr->show();
		    popupwin_ptr->write("Tempfile could not be opened", 2, 5);
		    popupwin_ptr->write("Check permissions in ", 3, 5);
		    popupwin_ptr->write(prog_func.filedir);
		    popupwin_ptr->winrefresh();
		    sleep(3);
		    popupwin_ptr->unshow();
		    delete popupwin_ptr;
		    refresh_winlist(*screenform_ptr);
		    tnc_ptr->tnc_func.capturefile_flag = FALSE;
		}

		do {
		    if (count) {  // terminate any former line with '\n'
		        tempfile.put('\n');
		    }
		    buffer[--extracted] = 0;
		    for (index = 0; (letter = buffer[index]) != 0
			     && count < MESSAGE_SIZE; index++, count++) {

		        tempfile.put(letter);
		    }

		} while (count < MESSAGE_SIZE && filein.getline(buffer, MESSAGE_SIZE) &&
			 *buffer != '#'  && *buffer != 0 && (extracted = filein.gcount()) > 0);
                tempfile.close();
		filein.close();
                sendfile(filename, TRUE);
		delete[] filename;
	    }
	}
	else {  // no entry in ./messages for this message number
	    beep();
	    doupdate();
	    filein.close();
	}
    }
    delete[] filename;
}

void Keyboard_input_with_data::sendfile(char* filename, int convert_codes) {
  // filename has a default value of 0 (null pointer), and convert_codes a default value of FALSE.
  // If filename is absent (filename == 0), a file in course of being sent will have more
  // of itself entered into tr_buffer.
  // If a filename is supplied, it will be opened, and convert_codes may be set to TRUE.
  // If if is set TRUE, then $a will be translated to your call, and $b will be 
  // translated to your selcall, and $c will be translated to the call in the callsign
  // store buffer for the currently active screen

  // if we were sending a file and now it has all gone to the Kam, signal this
    if (sendfile_flag == loading_complete && tnc_ptr->is_freebytes()) sendfile_flag = not_sending;

    char letter;
    int result = 0;
    int count;
    if (filename) {
        filein.open(filename, ios::in | ios::nocreate);
	if (!filein) {
	    PopupWin* popupwin_ptr = new PopupWin;
	    if (!popupwin_ptr) {
	        cerr << "Memory allocation error in Keyboard_controller::send_message()" << endl;
		exit(MEM_ERROR);
	    }
	    popupwin_ptr->show();
	    popupwin_ptr->write(filename, 1, 5);
	    popupwin_ptr->write("cannot be found or opened", 2, 5);
	    popupwin_ptr->winrefresh();
	    sleep(3);
	    popupwin_ptr->unshow();
	    delete popupwin_ptr;
	    refresh_winlist(*screenform_ptr);
	}
	else {
	    sendfile_flag = loading_buffer;
	    convert_codes_flag = convert_codes;
	}
    }

    for (count = 0; result != -1 && count < TR_BUFFER_SIZE/2 && filein.get(letter); count++) {
        if (tnc_ptr->tnc_func.active_port
	      && (tnc_ptr->tnc_func.hfmode == Tnc_func::rtty
		  || tnc_ptr->tnc_func.hfmode == Tnc_func::fec
		  || tnc_ptr->tnc_func.hfmode == Tnc_func::amtor
		  || (tnc_ptr->tnc_func.hfmode == Tnc_func::gtor
		      && tnc_ptr->tnc_func.stream_status[0][1] == Tnc_func::disconnected)
		  || (tnc_ptr->tnc_func.hfmode == Tnc_func::tor
			 && (tnc_ptr->tnc_func.stream_status[0][1] == Tnc_func::disconnected
			     || tnc_ptr->tnc_func.tor_connected_mode == Tnc_func::tor_amtor)))) {
	    if ((uchar)letter > 31) {
	        if (!is_baudot(letter)) letter = 0;
		else if (tnc_ptr->tnc_func.hfmode == Tnc_func::rtty) letter = toupper(letter);
	    }
	}


#if CHAR_SET==CP437
        if (((uchar) letter < 256 && (uchar)letter > 31) || letter == '\n') {
#elif CHAR_SET==LATIN-1
        if ((uchar)letter > 31 || letter == '\n') {
#else
	if ((letter < 127 && letter > 31) || letter == '\n') {
#endif
	    if (letter == '$' && convert_codes_flag == TRUE) {
	        int second_letter = filein.peek();
		if (second_letter == 'a') {
		    if ((unsigned int)tr_buffer_ptr->letters_free() >= strlen(prog_func.myCall)) {
		        for (char* letter_ptr = prog_func.myCall; *letter_ptr; letter_ptr++) {
			    tr_buffer_ptr->add_letter(*letter_ptr);
			    sendwin_ptr->write(*letter_ptr);
			}
			second_letter = filein.get(); // get rid of the 'a' code
		    }
		    else {           // not sufficient room in buffer for selcall
		        filein.putback('$');
			result = -1;
		    }
		}
		else if (second_letter == 'b') {
		    if ((unsigned int)tr_buffer_ptr->letters_free() >= strlen(prog_func.mySelCall)) {
		        for (char* letter_ptr = prog_func.mySelCall; *letter_ptr; letter_ptr++) {
			    tr_buffer_ptr->add_letter(*letter_ptr);
			    sendwin_ptr->write(*letter_ptr);
			}
			second_letter = filein.get(); // get rid of the 'b' code
		    }
		    else {           // not sufficient room in buffer for selcall
		        filein.putback('$');
			result = -1;
		    }
		}
		else if (second_letter == 'c') {
		    if ((unsigned int)tr_buffer_ptr->letters_free() >= 
			   strlen(tnc_ptr->tnc_func.hisCall[tnc_ptr->tnc_func.active_stream()][tnc_ptr->tnc_func.active_port])) {
		        for (char* letter_ptr = tnc_ptr->tnc_func.hisCall[tnc_ptr->tnc_func.active_stream()][tnc_ptr->tnc_func.active_port]; *letter_ptr; letter_ptr++) {
			    tr_buffer_ptr->add_letter(*letter_ptr);
			    sendwin_ptr->write(*letter_ptr);
			}
			second_letter = filein.get(); // get rid of the 'c' code
		    }
		    else {           // not sufficient room in buffer for selcall
		        filein.putback('$');
			result = -1;
		    }
		}
		else if (second_letter == EOF) {  // last letter of file is a '$'
		    result = tr_buffer_ptr->add_letter('$');
		    if (result == -1) filein.putback('$');
		    else {
		        sendwin_ptr->write('$');
			sendwin_ptr->winrefresh();
		    }
		}
	    }
	    else {
	        result = tr_buffer_ptr->add_letter(letter);
		if (result == -1) filein.putback(letter);
		else {
		    sendwin_ptr->write(letter);
		    if (letter == '\n' || letter == ' ') sendwin_ptr->winrefresh();
		}
	    }
	}
    }
    if (sendfile_flag == loading_buffer && !filein) {
        sendfile_flag = loading_complete;
	filein.close();
	convert_codes_flag = FALSE;
// make sure that if a temporary file was being used to send a message,it is deleted
        char* tempfile = new char[strlen(prog_func.filedir) + 11];
        if (!tempfile) {
            cerr << "Memory allocation error in Keyboard_controller.action()" << endl;
            exit(MEM_ERROR);
        }
        strcpy(tempfile, prog_func.filedir);
	strcat(tempfile, "/temp_file");
        unlink(tempfile);
        delete[] tempfile;
    }
    sendwin_ptr->winrefresh();
}

Keyboard_setup_mode::Keyboard_setup_mode(MainScreen& mainscreen, SendWin& sendwin,
                ReceiveWin& receivewin, WinList& screenform, Pipe_fifo& send_pipe,
		Transmit_buffer& tr_buffer, Tnc& tnc) {
    set_data(&mainscreen, &sendwin, &receivewin, &screenform, &send_pipe,
	         &tr_buffer, &tnc, this);
    sendfile_flag = not_sending;
    convert_codes_flag = FALSE;
    keyboard_ptr = this;
}

void Keyboard_setup_mode::action(void) {
    int termchar;
    int result;

    if ((termchar = getch()) != ERR) {  // non-blocking getch() returns ERR if no letter available to extract

        if ((termchar < 127 && termchar > 31) || termchar == '\n' || termchar == 3) {
	    result = tr_buffer_ptr->add_letter((char)termchar);
	    if (result == -1) beep();
	    else if (termchar != 3) {
	        sendwin_ptr->write((char)termchar);
		sendwin_ptr->winrefresh();
	    }
	}
	
	else if (termchar == KEY_BACKSPACE) {
	    result = tr_buffer_ptr->erase_letter();
	    if (!result) {
	        beep();
		doupdate();
	    }
	    else {
	        if (result == -1) {  // if last letter already extracted, we need to send value 8 to the Kam
		    do {
		        result = send_pipe_ptr->write(8);
		    } while (!result);
		}
		sendwin_ptr->write(8);        // now erase letter from screen
		sendwin_ptr->write(32);
		sendwin_ptr->write(8);
		sendwin_ptr->winrefresh();
	    }
	}
	else if (termchar == 17) {      // Ctrl-Q
	    exit_query_obj.setup();
	}
    }
}

void Gpl_query::action(void) {
    int letter = getch();
    if (letter == 'Y' || letter == 'y') {
        prog_func.GPL_flag = TRUE;
// GPL terms accepted - put a marker in the home directory
	char* gpl_file = new char[strlen(prog_func.homedir) + 9];
	if (!gpl_file) {
	    cerr << "Memory allocation error in Gpl_query::action()" << endl;
	    exit(MEM_ERROR);
	}
	strcpy(gpl_file, prog_func.homedir);
	strcat(gpl_file, "/");
	strcat(gpl_file, ".kamgpl");
	fileout.open(gpl_file, ios::out);
    }
    if (letter == 'N' || letter == 'n') {
        prog_func.exitflag = TRUE;
	prog_func.GPL_flag = TRUE;
    }
    else if (letter ==  KEY_UP) copyright_screen.scrollup();
    else if (letter == KEY_DOWN) copyright_screen.scrolldown();
}

int Keyboard_input_with_data::is_baudot(char letter) {
// this returns TRUE if the character falls within the permitted Baudot character set
// and otherwise it returns false.
// Both upper and lower case letters are passed as TRUE.
// Note that the US code substitutes ; for =, and " for +, as used by
// CCIT No. 2 code.  All of these are passed TRUE by is_baudot(), but may not be
// transmitted by the Kam, depending on the CODE RTTY and CODE AMTOR settings

    int return_value = FALSE;
    if (letter == '\n'
	|| (letter > 31 && letter < 64
	    && letter != 37 
	    && letter != 42
	    && letter != 60
	    && letter != 62)
	|| (letter > 64 && letter < 91)
	|| (letter > 96 && letter < 123)) {
        return_value = TRUE;
    }
    return return_value;
}
