/* Copyright (C) 1999 Chris Vine, G3XXF

This program is distributed under the General Public Licence, version 2.
For particulars of this and relevant disclaimers see the file
COPYRIGHT distributed with the source files.

*/

// tnc.cpp

#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include <ncurses.h>
// we don't want to use the 'clear' macro in ncurses.h - it will interfere with iostream::clear()
#undef clear
#include "tnc.h"
#include "keyboard.h"
#include "pipes.h"
#include "buffers.h"
#include "screen.h"


Tnc::Tnc(Transmit_buffer& a): Tnc_base(a), check_status_flag(0), info_received_flag(FALSE),
           rx_frame_complete(FALSE), kam_freebytes(0), unsent_frame_flag(FALSE),
           tx_frame_index(0), CMDflag(FALSE), tx_data(Tnc::tx_frame + 4),
	   CMDchar(0), de_flag(0), find_call_index(0) {
    rx_frame.command = rx_frame.port = rx_frame.stream = rx_frame.data[0] =
      rx_frame.data_count = rx_frame.FEND_count = rx_frame.bytes_afterFEND = 0;
    *callbuffer = 0;
    int port;
    int stream;
    for (port = 0; port < 2; port++) {
        for (stream = 0; stream < MAXUSERS; stream++) {
	    tnc_func.stream_status[stream][port] = Tnc_func::disconnected;
	    *(tnc_func.hisCall[stream][port]) = 0;
	}
    }
    *(tnc_func.selCall) = 0;
    tnc_func.hfmode = Tnc_func::packet;
    tnc_func.hisCall_lock = Tnc_func::off;
    tnc_func.speed_lock = Tnc_func::off;
    tnc_func.have_sent_lock = Tnc_func::no;
    tnc_func.active_hf_stream = tnc_func.active_vhf_stream = tnc_func.active_port = 0;
    tnc_func.capturefile_flag = FALSE;
    tnc_func.capture_stream.port = tnc_func.capture_stream.stream = 0;
}

void Tnc::process_received_byte(char modemchar) {
    if ((uchar)modemchar == FEND) { // we are either at the beginning or end of a frame
        if (!rx_frame.FEND_count || !rx_frame.bytes_afterFEND) { // we are at the beginning
	    rx_frame.FEND_count = 1;
	    rx_frame.bytes_afterFEND = 0;
	    rx_frame.data_count = 0;
	}
	else {                      // we are at the end
	    rx_frame_complete = TRUE;
	    rx_frame.FEND_count = 0;
	}
    }

    else if (!rx_frame.bytes_afterFEND && rx_frame.FEND_count) { // command byte
        rx_frame.command = modemchar;
	// if query reply frame, pretend we have received port and stream bytes
	if (modemchar == '?') {
	    rx_frame.bytes_afterFEND = 3; 
	    info_received_flag = TRUE;
	}
	else rx_frame.bytes_afterFEND++;
    }

    else if (rx_frame.bytes_afterFEND == 1) {   // port byte
        rx_frame.port = modemchar;
	rx_frame.bytes_afterFEND++;
    }
    
    else if (rx_frame.bytes_afterFEND == 2) {   // stream byte
        rx_frame.stream = modemchar;
	rx_frame.bytes_afterFEND++;
    }

    else if (rx_frame.FEND_count) {             // data byte
        rx_frame.data[rx_frame.data_count++] = modemchar;
    }
	    
    if (rx_frame_complete ) {
        rx_frame_complete = FALSE;
	if (rx_frame.data_count) {
	  process_rx_frame();
	}
    }
}

void Tnc::send_to_tnc(Pipe_fifo& send_pipe, int guard_flush) {

// if guard_flush is TRUE, then the requirement to buffer a number of characters
// equal to GUARD_QUEUE will be ignored

// send anything in the buffer or unsent frame
    if (tr_buffer.view_letter() != -1 || unsent_frame_flag) {
        int letter = 0;
        do {
	    if (!unsent_frame_flag) {  // make a new frame if none waiting to be sent
	        tx_frame[0] = FEND;
		tx_frame[1] = 'D';
		tx_frame[2] = tnc_func.active_port ? '2' : '1';
		
		char temp;
		if ((tnc_func.active_port && tnc_func.hfmode != Tnc_func::packet)
		      || tnc_func.stream_status[tnc_func.active_stream()]
		      [tnc_func.active_port] == Tnc_func::disconnected) {
		    temp = '0';
		}
		else temp = ((tnc_func.active_stream() | 0x40) + 1);
		tx_frame[3] = temp;   // now add the stream byte

		for (tx_frame_index = 0; CMDflag == FALSE
		       && tx_frame_index < FRAMEDATA_SIZE
		       && (guard_flush
			   || !tnc_func.active_port
			   || tnc_func.hfmode == Tnc_func::packet
			   || prog_func.send_mode != Prog_func::guard
			   || tr_buffer.letters_used() > GUARD_QUEUE)
		       && (letter = tr_buffer.extract_letter()) != -1;) { // this test must come last to prevent
		                                                          // incorrect abstraction from tr_buffer
		    if (letter == '\n') {
		        tx_data[tx_frame_index++] = '\r';
			guard_flush = FALSE;
		    }
		    else if ((uchar)letter == FEND) {
		        tx_data[tx_frame_index++] = FESC;
			tx_data[tx_frame_index++] = TFEND;
		    }
		    else if ((uchar)letter == FESC) {
		        tx_data[tx_frame_index++] = FESC;
			tx_data[tx_frame_index++] = TFESC;
		    }
    // check to see if there is a command in tr_buffer (the only command normally
    // placed in the buffer is the receive command 'E': others are placed directly
    // in send_pipe)
		    else if (letter == CMDinbuffer) {
		        CMDchar = tr_buffer.extract_letter();    // now get the command char itself
			tr_buffer.extract_letter();              // extract the second CMDinbuffer token
			CMDflag = TRUE;
		    }

#if CHAR_SET==LATIN-1
		    else if ((letter = Latin1_to_cp437(letter)) != 0) tx_data[tx_frame_index++] = letter;
#else
		    else tx_data[tx_frame_index++] = letter;
#endif
		}
		tx_frame[tx_frame_index + 4] = FEND;
	    }
	    else letter = 0;  // put dummy value in letter

	    
	    int result = send_pipe.write(tx_frame, tx_frame_index + 5);
	    if (!result) unsent_frame_flag = TRUE;
	    else {
	        unsent_frame_flag = FALSE;
		kam_freebytes -= tx_frame_index + 3;
	    }
	} while (unsent_frame_flag == FALSE
		   && CMDflag == FALSE
		   && kam_freebytes > FRAMEDATA_SIZE + 3
		   && (guard_flush
		       || !tnc_func.active_port
		       || tnc_func.hfmode == Tnc_func::packet
		       || prog_func.send_mode != Prog_func::guard
		       || tr_buffer.letters_used() > GUARD_QUEUE)
		   && letter != -1);
    }
    if (CMDflag == TRUE && !unsent_frame_flag) {          // now send any command
        CMDflag = FALSE;
	Keyboard_controller::send_specialcommand(CMDchar);
    }
}

int Tnc::is_freebytes(void) const {
    return (kam_freebytes > FRAMEDATA_SIZE + 3);
}

int Tnc::get_freebytes(void) const {
    return kam_freebytes;
}

void Tnc::process_rx_frame(void) {

    if (rx_frame.command == '?') {
        if (rx_frame.data[1] == 'H' || rx_frame.data[1] == 'J') { // "M" byte
	    tr_buffer.set_can_transmit_delete(TRUE);
	    tnc_func.tor_connected_mode = Tnc_func::tor_pactor_gtor;
	}
	else {
	    tr_buffer.set_can_transmit_delete(FALSE);
	    tnc_func.tor_connected_mode = Tnc_func::tor_amtor;
	}
	mainscreen_p->update_torinfo(rx_frame.data[3]); // "X" byte
	mainscreen_p->update_txinfo(rx_frame.data[4]);  // "Y" byte
	if (rx_frame.data[3] & 32) {                    // we are an ISS
	    tnc_func.have_sent_lock = Tnc_func::no;
	}
	else if (tnc_func.stream_status[0][1] == Tnc_func::connected
	      && tnc_func.speed_lock == Tnc_func::on
	      && tnc_func.have_sent_lock == Tnc_func::no) {
	    tnc_func.have_sent_lock = Tnc_func::yes;
	    Keyboard_controller::send_specialcommand('1');
	}
    }
    else {
        int port;
	int stream;
	int index;
        port = rx_frame.port & 0xf;  // remove upper nibble to convert ASCII to number
	if (port) port--;            // put in format for active_port
    
	if (rx_frame.stream == '0') stream = 0;
	else {
	    stream = rx_frame.stream & 0x1f; // remove top 3 bits to convert to number
	    if (stream) stream--;            // put in format for active_stream()
	}

	if (stream >= MAXUSERS) {
	    char message[] = "Too many streams received from Kam:\n"
	      "Check that MAXUSERS for Kam in tnc.parms matches the one in prog_defs.h\n";
	    for (index = 0; message[index]; index++) {
	        receivewin_p->ReceiveWin::write(message[index]);
		receivewin_p->store_letter(message[index]);
	    }
	    *rx_frame.data = 0;
	}
    
	else if (rx_frame.command == 'C') {  // buffer contains response to a command
	    if (check_status_flag) {  // we are checking free bytes in the Kam buffers
	        check_status_flag = 0;
		char* message;
		if (!(message = new char[rx_frame.data_count + 1])) {
		    cerr << "Memory allocation error in "
		      "process_rx_frame()" << endl;
		    exit(MEM_ERROR);
		}
		memcpy(message, rx_frame.data, rx_frame.data_count);
		message[rx_frame.data_count] = 0;  // null terminate the string
		parse_status(message);
		delete[] message;
		mainscreen_p->display_freebytes();
	    } 

	    else {    
	        for (index = 0; index < rx_frame.data_count; index++) {
		    if (rx_frame.data[index] < 127  && rx_frame.data[index] > 31) {
		        receivewin_p->ReceiveWin::write(rx_frame.data[index]);
		        receivewin_p->store_letter(rx_frame.data[index]);
		    }
		    else if (rx_frame.data[index] == '\r') {
		        receivewin_p->ReceiveWin::write('\n');
		        receivewin_p->store_letter('\n');
		    }
		}
	    }
	}

	else if (rx_frame.command == 'D') {   // buffer contains data
	    int stream_capture_flag = FALSE;
	    if (tnc_func.capturefile_flag &&
		tnc_func.capture_stream.stream == stream &&
		tnc_func.capture_stream.port == port) {
	        stream_capture_flag = TRUE;
	    }
	    for (index = 0; index < rx_frame.data_count; index++) {

#if CHAR_SET==CP437
	        if ((uchar)rx_frame.data[index] < 255  && (uchar)rx_frame.data[index] > 31) {
#elif CHAR_SET==LATIN-1
	        if ((uchar)rx_frame.data[index] < 255  && (uchar)rx_frame.data[index] > 31 &&
		    (rx_frame.data[index] = cp437_to_Latin1(rx_frame.data[index])) != 0) {
#else
	        if (rx_frame.data[index] < 127  && rx_frame.data[index] > 31) {
#endif

		    receivewin_p->store_letter(rx_frame.data[index], stream, port);
		    if (port == tnc_func.active_port
			&& stream == tnc_func.active_stream()) {
		        receivewin_p->ReceiveWin::write(rx_frame.data[index]);
		    }
		    if (stream_capture_flag) fileout.put(rx_frame.data[index]);
		}
		else if (rx_frame.data[index] == '\r') {
		    receivewin_p->store_letter('\n', stream, port);
		    if (port == tnc_func.active_port
			&& stream == tnc_func.active_stream()) {
		        receivewin_p->ReceiveWin::write('\n');
		    }
		    if (stream_capture_flag) fileout.put('\n');
		
		    if (!receivewin_p->is_scrolling()) {
		        receivewin_p->winrefresh();
		    }
		}

		else if (rx_frame.data[index] == 8) {
		    receivewin_p->unstore_letter(stream, port);
		    if (port == tnc_func.active_port
			&& stream == tnc_func.active_stream()) {
		        receivewin_p->ReceiveWin::write(8);
			receivewin_p->ReceiveWin::write(' ');
			receivewin_p->ReceiveWin::write(8);
		    }
		    if (stream_capture_flag) {
		        fileout.put(8);
			fileout.put(' ');
			fileout.put(8);
		    }
		    if (!receivewin_p->is_scrolling()) {
		        receivewin_p->winrefresh();
		    }
		}

		else if (rx_frame.data[index] == 7) {
		    beep();
		    doupdate();
		}
		if (tnc_func.hisCall_lock == Tnc_func::off && port
		    && tnc_func.hfmode != Tnc_func::packet) {
		    find_call(rx_frame.data[index]);
		}
	    }
	}
    
	else if (rx_frame.command == 'M') {// buffer contains unconnected packet data
	    int temp;
	    if (port) temp = tnc_func.active_hf_stream;
	    else temp = tnc_func.active_vhf_stream;
	    if (tnc_func.stream_status[temp][port] == Tnc_func::disconnected) {
	        for (index = 0; index < rx_frame.data_count; index++) {
		    if (rx_frame.data[index] < 127  && rx_frame.data[index] > 31) {
		        receivewin_p->store_letter(rx_frame.data[index], temp, port);
			if (port == tnc_func.active_port) {
			    receivewin_p->ReceiveWin::write(rx_frame.data[index]);
			}
		    }
		    else if (rx_frame.data[index] == '\r') {
		        receivewin_p->store_letter('\n', temp, port);
			if (port == tnc_func.active_port) {
			    receivewin_p->ReceiveWin::write('\n');
			    if (!receivewin_p->is_scrolling()) {
			        receivewin_p->winrefresh();
			    }
			}
		    }
		    else if (rx_frame.data[index] == 8) {
		        receivewin_p->unstore_letter(stream, port);
			if (port == tnc_func.active_port
			    && stream == tnc_func.active_stream()) {
			    receivewin_p->ReceiveWin::write(8);
			    receivewin_p->ReceiveWin::write(' ');
			    receivewin_p->ReceiveWin::write(8);
			}
			if (!receivewin_p->is_scrolling()) {
			  receivewin_p->winrefresh();
			}
		    }
		}
		// try to beautify the display of unconnected packets
		if (rx_frame.data[index - 1] == ':') {
		    receivewin_p->store_letter('\n', temp, port);
		    if (port == tnc_func.active_port) {
		        receivewin_p->ReceiveWin::write('\n');
			if (!receivewin_p->is_scrolling()) {
			  receivewin_p->winrefresh();
			}
		    }
		}
	    }
	}

	else if (rx_frame.command == 'E') {  // my transmitted data in HF mode
	    for (index = 0; index < rx_frame.data_count; index++) {
#if CHAR_SET==CP437
	        if ((uchar)rx_frame.data[index] < 255  && (uchar)rx_frame.data[index] > 31) {
#elif CHAR_SET==LATIN-1
	        if ((uchar)rx_frame.data[index] < 255  && (uchar)rx_frame.data[index] > 31 &&
		    (rx_frame.data[index] = cp437_to_Latin1(rx_frame.data[index])) != 0) {
#else
	        if (rx_frame.data[index] < 127  && rx_frame.data[index] > 31) {
#endif

		    receivewin_p->store_letter(rx_frame.data[index], 0, 1);
		    // next line no longer used
		    // tr_buffer.transmitted_letter(rx_frame.data[index]);
		    if (tnc_func.active_port == 1) {
		        receivewin_p->ReceiveWin::write(rx_frame.data[index]);
		    }
		}
		else if (rx_frame.data[index] == '\r') {
		    receivewin_p->store_letter('\n', 0, 1);
		    // next line no longer used
		    // tr_buffer.transmitted_letter('\n');
		    if (tnc_func.active_port == 1) {
		        receivewin_p->ReceiveWin::write('\n');
			if (!receivewin_p->is_scrolling()) {
			  receivewin_p->winrefresh();
			}
		    }
		}
		else if (rx_frame.data[index] == 8) {
		    receivewin_p->unstore_letter(0, 1);
		    if (tnc_func.active_port  == 1) {
		        receivewin_p->ReceiveWin::write(8);
			receivewin_p->ReceiveWin::write(' ');
			receivewin_p->ReceiveWin::write(8);
		    }
		    if (!receivewin_p->is_scrolling()) {
		        receivewin_p->winrefresh();
		    }
		}
	    }
	}

	else if (rx_frame.command == 'S') {   // buffer contains link state information
	    char* message;
	    if (!(message = new char[rx_frame.data_count + 1])) {
	        cerr << "Memory allocation error in "
		  "process_rx_frame(hostframe&, ReceiveWin&, MainScreen&, int&)" << endl;
		exit(MEM_ERROR);
	    }
	    for (index = 0; index < rx_frame.data_count; index++) {
	        if (rx_frame.data[index] < 127  && rx_frame.data[index] > 31) {
		    receivewin_p->store_letter(rx_frame.data[index], stream, port);
		    if (port == tnc_func.active_port
			&& stream == tnc_func.active_stream()) {
		        receivewin_p->ReceiveWin::write(rx_frame.data[index]);
		    }
		}
		else if (rx_frame.data[index] == '\r') {
		    receivewin_p->store_letter('\n', stream, port);
		    if (port == tnc_func.active_port
			&& stream == tnc_func.active_stream()) {
		        receivewin_p->ReceiveWin::write('\n');
		    }
		    if (!receivewin_p->is_scrolling()) {
		        receivewin_p->winrefresh();
		    }
		}
	    }
	    memcpy(message, rx_frame.data, rx_frame.data_count);
	    message[rx_frame.data_count] = 0; // make message a string
	    read_state_message(message, stream, port);
	    delete[] message;
	}
	if (!receivewin_p->is_scrolling()) {
	    receivewin_p->winrefresh();
	}
    }
}

void Tnc::read_state_message(char* message, int stream, int port) {
    char* message_ptr = 0;
    char* temp_ptr = 0;
    if (strstr(message, "*** DISCONNECTED")) {
        tnc_func.stream_status[stream][port] = Tnc_func::disconnected;
	mainscreen_p->display_connected_status();
        int count;
	for (count = 0; count < 3; count++) {
	    beep();
	    doupdate();
	    usleep(200000);
	}
    }
    else if ((message_ptr = strstr(message, "*** CONNECTED to ")) != 0) {
        message_ptr += 17;                  // update the callsign buffer
	// remove any trailing spaces (eg if CSTAMP is ON)
	for (temp_ptr = message_ptr; *temp_ptr && *temp_ptr != ' '; temp_ptr++) {}
	if (*temp_ptr == ' ') *temp_ptr = 0;
	
	strncpy(tnc_func.hisCall[stream][port], message_ptr, hisCall_SIZE);
	tnc_func.hisCall[stream][port][hisCall_SIZE] = 0;  // make sure the string is null terminated
	if (tnc_func.active_port && tnc_func.hfmode != Tnc_func::packet) {
	    tnc_func.hisCall_lock = Tnc_func::on;
	}
	mainscreen_p->display_callsign();
        tnc_func.stream_status[stream][port] = Tnc_func::connected;
	mainscreen_p->display_connected_status();
	int count;
	for (count = 0; count < 5; count++) {
	    beep();
	    doupdate();
	    usleep(200000);
	}
    }

    else if ((message_ptr = strstr(message, "<LINKED TO ")) != 0
	       && (temp_ptr = strchr(message, '>')) != 0) {
 // put the call-sign in the call buffer (for a non-Amtor link) in case we didn't initiate it
        if (tnc_func.hfmode != Tnc_func::amtor) {
	    *temp_ptr = 0;    // terminate the string at the ">"
	    message_ptr += 11;
	    strncpy(tnc_func.hisCall[0][1], message_ptr, hisCall_SIZE);// Tor mode is always stream 0
	    tnc_func.hisCall[0][1][hisCall_SIZE] = 0;  // make sure the string is null terminated
	    tnc_func.hisCall_lock = Tnc_func::on;
	    mainscreen_p->display_callsign();
	    make_selcall(message_ptr, tnc_func.selCall);
	}

        tnc_func.stream_status[0][1] = Tnc_func::connected;  
	mainscreen_p->display_connected_status();
	int count;
	for (count = 0; count < 5; count++) {
	    beep();
	    doupdate();
	    usleep(200000);
	}
    }
    
    else if (strstr(message, "<LINKED>")) { // Amtor link initiated by other station
        *(tnc_func.selCall) = 0; // we don't know the selcall (signal this)
	*(tnc_func.hisCall[0][1]) = 0; // or the callsign of the other station yet
	tnc_func.hisCall_lock = Tnc_func::off; // we can now pick the call-sign up
	mainscreen_p->display_callsign(); // blank the call-sign displayed

        tnc_func.stream_status[0][1] = Tnc_func::connected;  // Tor mode is always stream 0
	mainscreen_p->display_connected_status();
	int count;
	for (count = 0; count < 5; count++) {
	    beep();
	    doupdate();
	    usleep(200000);
	}
    }

    else if (strstr(message, "TOR STANDBY>")) {          // in Pactor, Amtor or Gtor
        tnc_func.stream_status[0][1] = Tnc_func::disconnected;  // Tor modes are always stream 0
	mainscreen_p->display_connected_status();
	if (tnc_func.hfmode == Tnc_func::tor) {
	    // go back to TOR mode (the Kam will have reset to the mode of the connecting
	    // station - we don't want this after that link finishes)
	
	    Keyboard_controller::send_specialcommand(packetCMD);
	    Keyboard_controller::send_kamcommand(torCMD);
	}
    }
    else if (strstr(message, "TOR_STANDBY>")             // in Kam TOR mode
	       && tnc_func.stream_status[0][1] == Tnc_func::connected) {
        tnc_func.stream_status[0][1] = Tnc_func::disconnected;  // Tor modes are always stream 0
	mainscreen_p->display_connected_status();
    }
}

void Tnc::parse_status(char* message) {
    int temp1 = 0;
    int temp2 = 0;
    char* temp_ptr = strchr(message, '\r'); // terminate the first line with a null
    if (temp_ptr) *temp_ptr = 0;
    else temp_ptr = message + strlen(message);
// if hfmode is Tor, we need to find the beginning of the second free byte number string
    if (tnc_func.hfmode != Tnc_func::packet) {
        temp_ptr = strchr(message, '/');
	if (temp_ptr) {
	  *temp_ptr = 0; // null terminate the first free byte number string
	  temp_ptr++;    // go to start of second free byte number string
	  temp1 = atoi(temp_ptr);
	}
    }
// now find the beginning of the first free byte number string
    if (temp_ptr) {
        for(; *temp_ptr != ' ' && temp_ptr > message; temp_ptr--);
	if (temp_ptr > message) { // check for sanity
	    temp_ptr++;    // go to start of first free byte number string
	    temp2 = atoi(temp_ptr);
	}
    }
    if (tnc_func.hfmode == Tnc_func::packet) kam_freebytes = temp2;
    else kam_freebytes = temp1 < temp2 ? temp1 : temp2;
}

void Tnc::find_freebytes(void) {
  // check_status_flag will prevent two concurrent requests
  // for a free_byte update, but will only block 5 times
  // to prevent locking

    if (!check_status_flag || !kam_freebytes) {
        check_status_flag = 5;
	Keyboard_controller::send_kamcommand(statusCMD);
    }
    else check_status_flag--;
}

void Tnc::get_info(void) {
    // if we are here because we have received an info frame
    // we don't want to ask for another
    if (info_received_flag) info_received_flag = FALSE;
    else if (tnc_func.hfmode != Tnc_func::packet) {
        Keyboard_controller::send_specialcommand('?');
    }
}

void Tnc::find_call(char modemchar) {
    char* message_ptr = 0;

    if (de_flag == 4) {
        if (find_call_index < CALLBUFFER_SIZE && modemchar != '\r' && modemchar != ' ') callbuffer[find_call_index++] = modemchar;
	else {
	    callbuffer[find_call_index] = 0;    // null terminate the string in callbuffer[]
	    find_call_index = 0;
	    de_flag = 0;
	    if ((message_ptr = strchr(callbuffer, ' ')) != 0) { // terminate string at first word space
	        *message_ptr = 0;
	    }
	    for(message_ptr = callbuffer; *message_ptr != 0; message_ptr++) {  // convert to upper case
	        *message_ptr = toupper(*message_ptr);
	    }
	    if (is_validcall(callbuffer)) {
	        strncpy(tnc_func.hisCall[0][1], callbuffer, hisCall_SIZE);
 // if an Amtor link is in progress which we did not initiate, lock the callsign
		if (tnc_func.stream_status[0][1] == Tnc_func::connected
		      && !*(tnc_func.selCall)) {
		    tnc_func.hisCall_lock = Tnc_func::on;
		}
		mainscreen_p->display_callsign();
 // now make up a selcall if we are not connected or we did not iniatate an Amtor link in progress
		if (tnc_func.stream_status[0][1] == Tnc_func::disconnected
		      || !*(tnc_func.selCall)) {
		    make_selcall(callbuffer, tnc_func.selCall);
		}
	    }
	}
    }

// look for " de " or " DE "

    if (modemchar == ' ' && de_flag < 3) de_flag = 1;
    else if (de_flag == 1  && (modemchar == 'd' || modemchar == 'D')) de_flag++;
    else if (de_flag == 2  && (modemchar == 'e' || modemchar == 'E')) de_flag++;
    else if (de_flag == 3  && modemchar == ' ') de_flag++;
    else if (de_flag < 4) {
        if (modemchar == '\r') de_flag = 1;
	else  de_flag = 0;
    }
}

int Tnc::is_validcall(const char* callsign) {  // returns TRUE if callsign valid, else returns FALSE
                                               // the contents of callsign must be in upper case
    char* char_ptr;
    if (!(char_ptr = new char[strlen(callsign) + 1])) {
        cerr << "Memory allocation error in int Tnc::is_validcall(char*)" << endl;
	exit(MEM_ERROR);
    }
    strcpy(char_ptr, callsign);
    char* temp_ptr;
    if ((temp_ptr = strchr(char_ptr, '/')) != 0) {  // split up the string if there is a '/'
        *temp_ptr = 0;
	temp_ptr++;   
	// temp_ptr now points at the right hand half, and char_ptr to the left hand half
	if (strlen(char_ptr) > strlen(temp_ptr)) temp_ptr = char_ptr; 
	// temp_ptr points to longer half now
    }
    else temp_ptr = char_ptr; 
        //  if no '/' in the callsign, temp_ptr points to a string holding callsign

    int invalid_call_flag = FALSE;
    temp_ptr++;  // move to second letter of main part of callsign
    if (*temp_ptr < '0' || *temp_ptr > '9') {  // second letter not number
        temp_ptr++;
	if (*temp_ptr < '0' || *temp_ptr > '9') {
	  // if second and third letters not numbers then invalid call
	    invalid_call_flag = TRUE;
	}
    }
    else if (*(temp_ptr - 1)  >= '0' && *(temp_ptr - 1) <= '9') { 
         // first and second letter is number - invalid call
        invalid_call_flag = TRUE;
    }
    else if (*(temp_ptr + 1)  >= '0' && *(temp_ptr + 1) <= '9') temp_ptr++;
         // second and third letter number - go to last number

    if (!invalid_call_flag) {
        temp_ptr++; // temp_ptr now points to the letter after the last or only number
        int num_letters = strlen(temp_ptr);
	if (num_letters  > 3 || !num_letters) {
         // if more than 3 letters or no letters after number, invalid call
	    invalid_call_flag = TRUE;
	}
	for (; *temp_ptr != 0; temp_ptr++) {
	    if (*temp_ptr < 'A' || *temp_ptr > 'Z') { // if any further numbers, invalid call
	        invalid_call_flag = TRUE;
	    }
	}
    }
    delete[] char_ptr;
    return (!invalid_call_flag);
}

void Tnc::make_selcall(const char* callsign,  char* selcall) {
  // the callsign should be in upper case
  // the selcall is copied into char* selcall
  
    char* char_ptr = new char[strlen(callsign) + 1];
    if (!char_ptr) {
        cerr << "Memory allocation error in Tnc::make_selcall(const char*, char*)" << endl;
	exit(MEM_ERROR);
    }
    strcpy(char_ptr, callsign);
    char* temp_ptr;
    if ((temp_ptr = strchr(char_ptr, '/')) != 0) {  // split up the string if there is a '/'
        *temp_ptr = 0;
	temp_ptr++;  
	  // temp_ptr now points at the right hand half, and char_ptr to the left hand half
	if (strlen(char_ptr) > strlen(temp_ptr)) temp_ptr = char_ptr;
	  // temp_ptr points to longer half now
    }
    else temp_ptr = char_ptr;
          //  if no '/' in the callsign, temp_ptr points to a string holding callsign
    
    char* temp_ptr2;
    for (temp_ptr2 = temp_ptr; *temp_ptr2 != 0; temp_ptr2++) {}; //find end of string
    temp_ptr2--;
    selcall[4] = 0;
    int lettercount = 3;
    while (lettercount > 0 && temp_ptr2 >= temp_ptr) {
        if (*temp_ptr2 >= 'A' && *temp_ptr2 <= 'Z') selcall[lettercount--] = *temp_ptr2;
	temp_ptr2--;
    }
          // find the first letter of the callsign (but stop at end of string)
    while (!(*temp_ptr == 0 || (*temp_ptr >= 'A' && *temp_ptr <= 'Z'))) temp_ptr++;
          // now fill in rest of selcall
    for(; lettercount >= 0; lettercount--) selcall[lettercount] = *temp_ptr;
    delete[] char_ptr;
}

uchar Tnc::cp437_to_Latin1(uchar letter) {
    if (letter > 127) {
        switch(letter) {
	    case 0x80: letter = 0xc7; break;
	    case 0x81: letter = 0xfc; break;
	    case 0x82: letter = 0xe9; break;
	    case 0x83: letter = 0xe2; break;
	    case 0x84: letter = 0xe4; break;
	    case 0x85: letter = 0xe0; break;
	    case 0x86: letter = 0xe5; break;
	    case 0x87: letter = 0xe7; break;
	    case 0x88: letter = 0xea; break;
	    case 0x89: letter = 0xeb; break;
	    case 0x8a: letter = 0xe8; break;
	    case 0x8b: letter = 0xef; break;
	    case 0x8c: letter = 0xee; break;
	    case 0x8d: letter = 0xec; break;
	    case 0x8e: letter = 0xc4; break;
	    case 0x8f: letter = 0xc5; break;

	    case 0x90: letter = 0xc9; break;
	    case 0x91: letter = 0xe6; break;
	    case 0x92: letter = 0xc6; break;
	    case 0x93: letter = 0xf4; break;
	    case 0x94: letter = 0xf6; break;
	    case 0x95: letter = 0xf2; break;
	    case 0x96: letter = 0xfb; break;
	    case 0x97: letter = 0xf9; break;
	    case 0x98: letter = 0xff; break;
	    case 0x99: letter = 0xd6; break;
	    case 0x9a: letter = 0xdc; break;
	    case 0x9b: letter = 0xa2; break;
	    case 0x9c: letter = 0xa3; break;
	    case 0x9d: letter = 0xa5; break;

	    case 0xa0: letter = 0xe1; break;
	    case 0xa1: letter = 0xed; break;
	    case 0xa2: letter = 0xf3; break;
	    case 0xa3: letter = 0xfa; break;
	    case 0xa4: letter = 0xf1; break;
	    case 0xa5: letter = 0xd1; break;
	    case 0xa8: letter = 0xbf; break;
	    case 0xab: letter = 0xbd; break;
	    case 0xac: letter = 0xbc; break;
	    case 0xad: letter = 0xa1; break;
	    case 0xae: letter = 0xab; break;
	    case 0xaf: letter = 0xbb; break;

	    case 0xe1: letter = 0xdf; break;
	    case 0xe6: letter = 0xb5; break;

	    case 0xf1: letter = 0xb1; break;
	    case 0xf6: letter = 0xf7; break;
	    case 0xf8: letter = 0xb0; break;
	    case 0xf9: case 0xfa: letter = 0xb7; break;
	    case 0xfd: letter = 0xb2; break;

	    default: letter = 0;
	}
    }
    return letter;
}

uchar Tnc::Latin1_to_cp437(uchar letter) {
    if (letter > 127) {
        switch(letter) {
	    case 0xa1: letter = 0xad; break;
	    case 0xa2: letter = 0x9b; break;
	    case 0xa3: letter = 0x9c; break;
	    case 0xa5: letter = 0x9d; break;
	    case 0xab: letter = 0xae; break;

	    case 0xb0: letter = 0xf8; break;
	    case 0xb1: letter = 0xf1; break;
	    case 0xb2: letter = 0xfd; break;
	    case 0xb5: letter = 0xe6; break;
	    case 0xb7: letter = 0xf9; break;
	    case 0xbb: letter = 0xaf; break;
	    case 0xbc: letter = 0xac; break;
	    case 0xbd: letter = 0xab; break;
	    case 0xbf: letter = 0xa8; break;

	    case 0xc4: letter = 0x8e; break;
	    case 0xc5: letter = 0x8f; break;
	    case 0xc6: letter = 0x92; break;
	    case 0xc7: letter = 0x80; break;
	    case 0xc9: letter = 0x90; break;

	    case 0xd1: letter = 0xa5; break;
	    case 0xd6: letter = 0x99; break;
	    case 0xdc: letter = 0x9a; break;
	    case 0xdf: letter = 0xe1; break;

	    case 0xe0: letter = 0x85; break;
	    case 0xe1: letter = 0xa0; break;
	    case 0xe2: letter = 0x83; break;
	    case 0xe4: letter = 0x84; break;
	    case 0xe5: letter = 0x86; break;
	    case 0xe6: letter = 0x91; break;
	    case 0xe7: letter = 0x87; break;
	    case 0xe8: letter = 0x8a; break;
	    case 0xe9: letter = 0x82; break;
	    case 0xea: letter = 0x88; break;
	    case 0xeb: letter = 0x89; break;
	    case 0xec: letter = 0x8d; break;
	    case 0xed: letter = 0xa1; break;
	    case 0xee: letter = 0x8c; break;
	    case 0xef: letter = 0x8b; break;

	    case 0xf1: letter = 0xa4; break;
	    case 0xf2: letter = 0x95; break;
	    case 0xf3: letter = 0xa2; break;
	    case 0xf4: letter = 0x93; break;
	    case 0xf6: letter = 0x94; break;
	    case 0xf7: letter = 0xf6; break;
	    case 0xf9: letter = 0x97; break;
	    case 0xfa: letter = 0xa3; break;
	    case 0xfb: letter = 0x96; break;
	    case 0xfc: letter = 0x81; break;
	    case 0xff: letter = 0x98; break;

	    default: letter = 0;
	}
    }
    return letter;
}

void Tnc_setup::process_received_byte(char modemchar) {
    if (modemchar < 127  && modemchar > 31) {
        receivewin_p->ReceiveWin::write(modemchar);
    }
    else if (modemchar == '\r') {
        receivewin_p->ReceiveWin::write('\n');
    }
    receivewin_p->winrefresh();
}

void Tnc_setup::send_to_tnc(Pipe_fifo& send_pipe, int letter) {

  // we don't need to use the value of letter passed to the function
  // because it is a dummy in Tnc_setup, so we can use it as a general
  // local variable (this also stops unnecesary compiler warnings)

    if ((letter = tr_buffer.extract_letter()) != -1) {
        if (letter == '\n') letter = '\r';
	send_pipe.write(letter);
    }
}

int Tnc_setup::is_freebytes(void) const {
    return TRUE;
}

int Tnc_setup::get_freebytes(void) const {
    return 1;
}

void Tnc_setup::find_freebytes(void) {}

void Tnc_setup::get_info(void) {}
